@@ -1,5 +1,5 @@
 /* This file is part of GNU Rush.                  
-   Copyright (C) 2008-2010, 2014 Sergey Poznyakoff
+   Copyright (C) 2008-2019 Sergey Poznyakoff
 
    GNU Rush is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -23,9 +23,9 @@ int lint_option = 0;
 unsigned sleep_time = 5;
 unsigned debug_level;
 int debug_option;
+char *dump_option;
 struct rush_rule *rule_head, *rule_tail;
 struct passwd *rush_pw;
-char *rush_interactive_shell;
 
 #define STDOUT_FILENO 1
 #define STDERR_FILENO 2
@@ -64,32 +64,33 @@ int
 string_to_error_index(const char *name)
 {
 	static const char *error_msg_name[] = {
-		"usage-error",
-		"nologin-error",
-		"config-error",
-		"system-error",
+		[usage_error]   = "usage-error",
+		[nologin_error] = "nologin-error",
+		[config_error]  = "config-error",
+		[system_error]  = "system-error",
 		NULL
 	};
-	static int error_msg_index[] = {
-		usage_error,
-		nologin_error,
-		config_error,
-		system_error
-	};
-	ARGMATCH_VERIFY(error_msg_name, error_msg_index);
+	int i;
 
-	ptrdiff_t d = ARGMATCH(name, error_msg_name, error_msg_index);
-	if (d < 0)
-		return -1;
-	return error_msg_index[d];
+	for (i = 0; error_msg_name[i]; i++)
+		if (strcmp(error_msg_name[i], name) == 0)
+			return i;
+	return -1;
 }
 	
 
 void
 send_msg(const char *msg, size_t len)
 {
-        if (write(STDERR_FILENO, msg, len) < 0)
-                write(STDOUT_FILENO, msg, len);
+        if (write(STDERR_FILENO, msg, len) < 0) {
+		logmsg(LOG_ERR,
+		       _("failed to write message to stderr: %s"),
+		       strerror(errno));
+		if (write(STDOUT_FILENO, msg, len) < 0)
+			logmsg(LOG_ERR,
+			       _("failed to write message to stdout: %s"),
+			       strerror(errno));
+	}
 }
 
 void
@@ -162,56 +163,46 @@ xalloc_die()
 {
         die(system_error, NULL, _("Not enough memory"));
 }
-
-int
-is_prefix(const char *pref, const char *str)
-{
-        int len = strlen(pref);
-        int slen = strlen(str);
-        
-        if (slen < len)
-                return 0;
-        if (memcmp(str, pref, len))
-                return 0;
-        if (str[len] != '/')
-                return 0;
-        return 1;
-}
-
-int
-is_suffix(const char *suf, const char *str)
-{
-        int len = strlen(suf);
-        int slen = strlen(str);
-        
-        if (slen < len)
-                return 0;
-        if (memcmp(str + slen - len, suf, len))
-                return 0;
-        if (slen > len && str[slen - len - 1] != '/')
-                return 0;
-        return 1;
-}
-
 
-int
+static int
+test_regex(struct rush_request *req, regex_t *rx, char const *subj)
+{
+	int rc;
+	struct rush_backref *bref = &req->backref[!req->backref_cur];
+	size_t n = rx->re_nsub + 1;
+	if (n > bref->maxmatch) {
+		bref->match = xrealloc(bref->match,
+				       sizeof(bref->match[0]) * n);
+		bref->maxmatch = n;
+	}
+	rc = regexec(rx, subj, bref->maxmatch, bref->match, 0);
+	if (rc == 0) {
+		free(bref->subject);
+		bref->subject = xstrdup(subj);
+		bref->nmatch = n;
+		req->backref_cur = !req->backref_cur;
+	}
+	return rc;
+}
+
+static int
 test_request_cmdline(struct test_node *node, struct rush_request *req)
 {
-        return regexec(&node->v.regex, req->cmdline, 0, NULL, 0);
+	return test_regex(req, &node->v.regex, req->cmdline);
 }
 
 #define ARG_NO(n,argc) (((n) < 0) ? (argc) + (n) : (n))
 
-int
+static int
 test_request_arg(struct test_node *node, struct rush_request *req)
 {
         int n = ARG_NO(node->v.arg.arg_no, req->argc);
         if (n < 0 || n >= req->argc)
                 return 1;
-        return regexec(&node->v.arg.regex, req->argv[n], 0, NULL, 0);
+        return test_regex(req, &node->v.arg.regex, req->argv[n]);
 }
 
-int
+static int
 test_num_p(struct test_numeric_node *node, unsigned long val)
 {
         switch (node->op) {
@@ -231,25 +222,25 @@ test_num_p(struct test_numeric_node *node, unsigned long val)
         return 0;
 }
 
-int
+static int
 test_request_argc(struct test_node *node, struct rush_request *req)
 {
         return !test_num_p(&node->v.num, req->argc);
 }
 
-int
+static int
 test_request_uid(struct test_node *node, struct rush_request *req)
 {
         return !test_num_p(&node->v.num, req->pw->pw_uid);
 }
 
-int
+static int
 test_request_gid(struct test_node *node, struct rush_request *req)
 {
         return !test_num_p(&node->v.num, req->pw->pw_gid);
 }
 
-int
+static int
 groupcmp(char *gname, struct passwd *pw)
 {
         struct group *grp;
@@ -268,7 +259,7 @@ groupcmp(char *gname, struct passwd *pw)
         return 1;
 }
 
-int
+static int
 test_request_group(struct test_node *node, struct rush_request *req)
 {
         char **p;
@@ -279,7 +270,7 @@ test_request_group(struct test_node *node, struct rush_request *req)
         return 1;
 }
 
-int
+static int
 test_request_user(struct test_node *node, struct rush_request *req)
 {
         char **p;
@@ -290,17 +281,17 @@ test_request_user(struct test_node *node, struct rush_request *req)
         return 1;
 }
 
-int (*test_request[])(struct test_node *, struct rush_request *) = {
-        test_request_cmdline,
-        test_request_arg,
-        test_request_argc,
-        test_request_uid,
-        test_request_gid,
-        test_request_user,
-        test_request_group
+static int (*test_request[])(struct test_node *, struct rush_request *) = {
+        [test_cmdline]     = test_request_cmdline,
+        [test_arg]         = test_request_arg,
+        [test_argc]        = test_request_argc,
+        [test_uid]         = test_request_uid,
+        [test_gid]         = test_request_gid,
+        [test_user]        = test_request_user,
+        [test_group]       = test_request_group,
 };
 
-int
+static int
 run_tests(struct rush_rule *rule, struct rush_request *req)
 {
         struct test_node *node;
@@ -312,6 +303,7 @@ run_tests(struct rush_rule *rule, struct rush_request *req)
 			    &req->i18n,
                             _("%s:%d: INTERNAL ERROR: node type out of range"),
                             __FILE__, __LINE__);
+
                 res = test_request[node->type](node, req);
                 if (node->negate)
                         res = !res;
@@ -321,16 +313,6 @@ run_tests(struct rush_rule *rule, struct rush_request *req)
         return 0;
 }
 
-struct rush_rule *
-match_rule(struct rush_rule *rule, struct rush_request *req)
-{
-        for (; rule; rule = rule->next) {
-                if (run_tests(rule, req) == 0)
-                        break;
-        }
-        return rule;
-}
-
 char *
 make_file_name(const char *dir, const char *name)
 {
@@ -364,22 +346,58 @@ expand_tilde(const char *dir, const char *home)
         return res;
 }
 
-static char *
-find_env(char *name, int val)
+char *
+expand_dir(const char *dir, struct rush_request *req)
 {
-        int nlen = strcspn(name, "?+=");
+	char *exp = rush_expand_string(dir, req);
+	if (exp[0] == '~') {
+		char *t = expand_tilde(exp, req->pw->pw_dir);
+		free(exp);
+		exp = t;
+	}
+	return exp;
+}
+
+/* Find variable NAME in environment ENV.
+   On success, store the index of the ENV slot in *IDX,
+   the offset of the value (position right past '=') in *VALOFF, and
+   return 0 (IDX and/or VALOFF can be NULL, if that info is not needed).
+   Return -1 if NAME was not found. */
+static int
+find_env_pos(char **env, char *name, int *idx, int *valoff)
+{
+        int nlen = strcspn(name, "+=");
         int i;
 
-        for (i = 0; environ[i]; i++) {
-                size_t elen = strcspn(environ[i], "=");
-                if (elen == nlen && memcmp(name, environ[i], nlen) == 0)
-                        return val ? environ[i] + elen + 1 : environ[i];
+        for (i = 0; env[i]; i++) {
+                size_t elen = strcspn(env[i], "=");
+                if (elen == nlen && memcmp(name, env[i], nlen) == 0) {
+			if (idx)
+				*idx = i;
+			if (valoff)
+				*valoff = elen + 1;
+			return 0;
+		}
         }
-        return NULL;
+        return -1;
 }
 
+/* Find variable NAME in environment ENV.
+   On success, return pointer to the variable assignment (if VAL is 0),
+   or to the value (if VAL is 1).
+   Return NULL if NAME is not present in ENV. */
+static char *
+find_env_ptr(char **env, char *name, int val)
+{
+	int i, j;
+	if (find_env_pos(env, name, &i, &j))
+		return NULL;
+	return val ? env[i] + j : env[i];
+}
+
+/* Return 1 if ENV contains a matching unset statement for variable NAME. */
 static int
-locate_unset(char **env, const char *name)
+var_is_unset(char **env, const char *name)
 {
         volatile int i;
         int nlen = strcspn(name, "=");
@@ -409,7 +427,7 @@ env_concat(char *name, size_t namelen, char *a, char *b)
         if (a && b) {
                 res = xmalloc(namelen + 1 + strlen(a) + strlen(b) + 1);
                 strcpy(res + namelen + 1, a);
-                strcat(res, b);
+                strcat(res + namelen + 1, b);
         } else if (a) {
                 len = strlen(a);
                 if (c_ispunct(a[len-1]))
@@ -420,6 +438,7 @@ env_concat(char *name, size_t namelen, char *a, char *b)
         } else /* if (a == NULL) */ {
                 if (c_ispunct(b[0]))
                         b++;
+		len = strlen(b);
                 res = xmalloc(namelen + 1 + len + 1);
                 strcpy(res + namelen + 1, b);
         }
@@ -433,7 +452,7 @@ env_setup(char **env)
 {
         char **old_env = environ;
         char **new_env;
-        int count, i, n;
+        int count, i, j, n;
         
         if (!env)
                 return old_env;
@@ -460,67 +479,70 @@ env_setup(char **env)
         
         if (old_env)
                 for (i = 0; old_env[i]; i++) {
-                        if (!locate_unset(env, old_env[i]))
+                        if (!var_is_unset(env, old_env[i]))
                                 new_env[n++] = old_env[i];
                 }
 
         for (i = 0; env[i]; i++) {
                 char *p;
                 
-                if (env[i][0] == '-') {
+                if (env[i][0] == '-')
                         /* Skip unset directives. */
                         continue;
-                } if ((p = strchr(env[i], '='))) {
+
+		/* Find the slot for the variable.  Use next available
+		   slot if there's no such variable in new_env */
+		if (find_env_pos(new_env, env[i], &j, NULL))
+			j = n;
+		
+		if ((p = strchr(env[i], '='))) {
                         if (p == env[i])
                                 continue; /* Ignore erroneous entry */
-                        if (p[-1] == '+') 
-                                new_env[n++] = env_concat(env[i],
-                                                          p - env[i] - 1,
-                                                          find_env(env[i], 1),
-                                                          p + 1);
-                        else if (p[1] == '+')
-                                new_env[n++] = env_concat(env[i],
-                                                          p - env[i],
-                                                          p + 2,
-                                                          find_env(env[i], 1));
-			else if (p[-1] == '?') {
-				if (!find_env(env[i], 0))
-					new_env[n++] = p + 1;
+                        if (p[-1] == '+') {
+				new_env[j] = env_concat(env[i],
+							p - env[i] - 1,
+							find_env_ptr(environ,
+								     env[i],
+								     1),
+							p + 1);
+			} else if (p[1] == '+') {
+                                new_env[j] = env_concat(env[i],
+							p - env[i],
+							p + 2,
+							find_env_ptr(environ,
+								     env[i],
+								     1));
 			} else
-                                new_env[n++] = env[i];
-                } else {
-                        p = find_env(env[i], 0);
-                        if (p)
-                                new_env[n++] = p;
-                }
+                                new_env[j] = env[i];
+                } else if ((p = find_env_ptr(environ, env[i], 0)))
+			new_env[j] = p;
+		else
+			continue;
+		/* Adjust environment size */
+		if (j == n)
+			++n;
         }
         new_env[n] = NULL;
         return new_env;
 }
-
-void
+
+static void
 reparse_cmdline(struct rush_request *req)
 {
-        int rc;
         struct wordsplit ws;
 	
         argcv_free(req->argc, req->argv);
 	if (wordsplit(req->cmdline, &ws, WRDSF_DEFFLAGS))
 		die(system_error, &req->i18n, _("wordsplit(%s) failed: %s"),
                     req->cmdline, wordsplit_strerror(&ws));
-	req->argc = ws.ws_wordc;
-	req->argv = ws.ws_wordv;
-
-	ws.ws_wordc = 0;
-	ws.ws_wordv = NULL;
-
+	wordsplit_get_words(&ws, &req->argc, &req->argv);
 	wordsplit_free(&ws);
 
 	free(req->prog);
 	req->prog = NULL;
 }
 
-void
+static void
 rebuild_cmdline(struct rush_request *req)
 {
         free(req->cmdline);
@@ -543,118 +565,243 @@ get_arg_no(int index, struct rush_request *req)
 static void
 assign_string(char **pstr, char *val)
 {
-	debug2(2, _("Transform: \"%s\" -> \"%s\""), *pstr ? *pstr : "", val);
+	debug(2, _("Transform: \"%s\" -> \"%s\""), *pstr ? *pstr : "", val);
 	free(*pstr);
 	*pstr = val;
 }
+
+static int
+transform_cmdline_fun(struct rush_request *req, struct transform_node *node,
+		      char *val, char **return_val)
+{
+	char *p;
+	
+	debug(2, "%s", _("Transforming command line"));
+	if (node->pattern) {
+		char *val = rush_expand_string(node->pattern, req);
+		p = transform_string(node->v.trans, val);
+		free(val);
+	} else
+		p = transform_string(node->v.trans, req->cmdline);
+	assign_string(&req->cmdline, p);
+	debug(2, _("Command line: %s"), req->cmdline);
+	return 0;
+}
 
-void
+static int
+transform_setcmd_fun(struct rush_request *req, struct transform_node *node,
+		     char *val, char **return_val)
+{
+	debug(2, "%s", _("Setting command line"));
+	assign_string(&req->cmdline, xstrdup(val));
+	debug(2, _("Command line: %s"), req->cmdline);
+	return 0;
+}
+
+static int
+transform_arg_fun(struct rush_request *req, struct transform_node *node,
+		  char *val, char **return_val)
+{
+	char *p = transform_string(node->v.trans, val);
+	assign_string(return_val, p);
+	return 1;
+}
+
+static int
+transform_map_fun(struct rush_request *req, struct transform_node *node,
+		  char *val, char **return_val)
+{
+	char *p;
+	debug(2,
+	      _("Transformation map: %s, %s, %s, %u, %u, %s"),
+	      node->v.map.file,
+	      node->v.map.delim,
+	      node->v.map.key,
+	      node->v.map.key_field,
+	      node->v.map.val_field,
+	      node->v.map.defval);
+	p = map_string(&node->v.map, req);
+	if (p) {
+		assign_string(return_val, p);
+		return 1;
+	}
+	return 0;
+}
+
+static int
+transform_delarg_fun(struct rush_request *req, struct transform_node *node,
+		     char *val, char **return_val)
+{
+	int i, arg_no, arg_end;
+	
+	arg_no = get_arg_no(node->arg_no, req);
+	arg_end = get_arg_no(node->v.arg_end, req);
+	if (arg_end < arg_no) {
+		int x = arg_end;
+		arg_end = arg_no;
+		arg_no = x;
+	}
+	debug(2, _("Deleting arguments %d-%d"), arg_no, arg_end);
+	if (arg_no == 0 || arg_end == 0)
+		die(config_error,
+		    &req->i18n, _("Deleting argv[0] is prohibited"));
+	for (i = arg_no; i <= arg_end; i++) 
+		free(req->argv[i]);
+	i = arg_end - arg_no + 1;
+	memmove(req->argv + arg_no,
+		req->argv + arg_end + 1,
+		(req->argc - i) * sizeof(req->argv[0]));
+	req->argc -= i;
+	return 1;
+}
+
+static int
+transform_setarg_fun(struct rush_request *req, struct transform_node *node,
+		     char *val, char **return_val)
+{
+	assign_string(return_val, xstrdup(val));
+	return 1;
+}
+
+static int
+transform_setvar_fun(struct rush_request *req, struct transform_node *node,
+		     char *val, char **return_val)
+{
+	size_t i;
+
+	if (req->var_kv) {
+		for (i = 0; i < req->var_count; i += 2)
+			if (strcmp(req->var_kv[i], node->v.varname) == 0)
+				break;
+	} else
+		i = req->var_count;
+
+	if (i < req->var_count) {
+		free(req->var_kv[i + 1]);
+		if (val)
+			req->var_kv[i + 1] = xstrdup(val);
+		else {
+			free(req->var_kv[i]);
+			memmove(req->var_kv + i, req->var_kv + i + 2,
+				(req->var_count - i - 1) * sizeof(req->var_kv[0]));
+			req->var_count -= 2;
+		}
+	} else if (val) {
+		while (req->var_count + 3 >= req->var_max)
+			req->var_kv = x2nrealloc(req->var_kv, &req->var_max,
+						 sizeof(req->var_kv[0]));
+		req->var_kv[req->var_count++] = xstrdup(node->v.varname);
+		req->var_kv[req->var_count++] = xstrdup(val);
+		req->var_kv[req->var_count] = NULL;
+	}
+	return 1;
+}
+
+/* Transform flags */
+#define XFORM_DFL         0x00 /* Default: nothing */
+#define XFORM_CMDLINE     0x01 /* Function operates on entire command line */
+#define XFORM_VALUE       0x02 /* Function needs value */
+#define XFORM_CHARGV      0x04 /* Function can change argv */
+
+struct transform_function
+{
+	int flags;
+	int (*func)(struct rush_request *req, struct transform_node *node,
+		    char *val, char **return_val);
+};
+
+static struct transform_function transform_funtab[] = {
+	[transform_cmdline] = {
+		XFORM_CMDLINE,
+		transform_cmdline_fun
+	},
+	[transform_setcmd]  = {
+		XFORM_CMDLINE|XFORM_VALUE,
+		transform_setcmd_fun
+	},
+	[transform_arg]     = {
+		XFORM_VALUE,
+		transform_arg_fun
+	},
+	[transform_map]     = {
+		XFORM_VALUE,
+		transform_map_fun
+	},
+	[transform_delarg]  = {
+		XFORM_CHARGV,
+		transform_delarg_fun
+	},
+	[transform_setarg]  = {
+		XFORM_VALUE,
+		transform_setarg_fun
+	},
+	[transform_setvar] = {
+		XFORM_VALUE,
+		transform_setvar_fun
+	},
+	[transform_unsetvar] = {
+		XFORM_DFL,
+		transform_setvar_fun
+	}		
+};	
+static int transform_count =
+	sizeof(transform_funtab)/sizeof(transform_funtab[0]);
+
+static void
 run_transforms(struct rush_rule *rule, struct rush_request *req)
 {
         struct transform_node *node;
-        char *p;
-        int i, arg_no, arg_end;
-        int args_transformed = 0;
 	char *val, **target;
+	char *mem = NULL;
+        int args_transformed = 0;
+	int res;
+	int flags;
 	
-#define GET_TGT_VAL()							\
-	if (node->progmod) {						\
-		arg_no = 0;						\
-		target = &req->prog;					\
-		val = PROGFILE(req);					\
-		debug1(2, _("Modifying program name (%s)"), val);	\
-	} else {							\
-		arg_no = get_arg_no(node->arg_no, req);			\
-		target = &req->argv[arg_no];				\
-		val = *target;						\
-		args_transformed = 1;					\
-		debug1(2, _("Modifying argv[%d]"), arg_no);		\
-	} 								\
-	if (node->pattern) {						\
-		val = rush_expand_string(node->pattern, req);		\
-	}
-
         for (node = rule->transform_head; node; node = node->next) {
-                switch (node->type) {
-                case transform_cmdline:
-                        if (args_transformed) {
-                                rebuild_cmdline(req);
-                                args_transformed = 0;
-                        }
-                        debug(2, _("Transforming command line"));
-                        p = transform_string(node->v.trans, req->cmdline);
-                        free(req->cmdline);
-                        req->cmdline = p;
-                        debug1(2, _("Command line: %s"), req->cmdline);
-                        reparse_cmdline(req);
-                        break;
+		if (node->type < 0 || node->type >= transform_count)
+			die(system_error, &req->i18n,
+			    _("%s:%d: internal error"), __FILE__, __LINE__);
 
-		case transform_setcmd:
-			GET_TGT_VAL();
-			if (args_transformed) {
-                                rebuild_cmdline(req);
-                                args_transformed = 0;
-                        }
-                        debug(2, _("Setting command line"));
-                        free(req->cmdline);
-                        req->cmdline = xstrdup(val);
-                        debug1(2, _("Command line: %s"), req->cmdline);
-                        reparse_cmdline(req);
-                        break;
-			
-                case transform_arg:
-			GET_TGT_VAL();
-                        p = transform_string(node->v.trans, val);
-			assign_string(target, p);
-			break;
-
-		case transform_map:
-			GET_TGT_VAL();
-			debug6(2,
-			       _("Transformation map: %s, %s, %s, %u, %u, %s"),
-			       node->v.map.file,
-			       node->v.map.delim,
-			       node->v.map.key,
-			       node->v.map.key_field,
-			       node->v.map.val_field,
-			       node->v.map.defval);
-                        p = map_string(&node->v.map, req);
-			if (p) 
-				assign_string(target, p);
-			else
-				args_transformed = 0;
-			break;
-
-		case transform_delarg:
-			arg_no = get_arg_no(node->arg_no, req);
-			arg_end = get_arg_no(node->v.arg_end, req);
-			if (arg_end < arg_no) {
-				int x = arg_end;
-				arg_end = arg_no;
-				arg_no = x;
+		val = NULL;
+		target = NULL;
+		flags = transform_funtab[node->type].flags;
+		
+		if ((flags & XFORM_CMDLINE) && args_transformed) {
+			rebuild_cmdline(req);
+			args_transformed = 0;
+		}
+		
+		if (flags & XFORM_VALUE) {
+			if (node->progmod) {
+				target = &req->prog;
+				val = PROGFILE(req);
+				debug(2, _("Modifying program name (%s)"), val);
+				flags &= ~XFORM_CHARGV;
+			} else {
+				int arg_no = get_arg_no(node->arg_no, req);
+				target = &req->argv[arg_no];
+				val = *target;
+				flags |= XFORM_CHARGV;
+				debug(2, _("Modifying argv[%d]"), arg_no);
 			}
-			debug2(2, _("Deleting arguments %d-%d"),
-			       arg_no, arg_end);
-			if (arg_no == 0 || arg_end == 0)
-				die(config_error,
-				    &req->i18n, _("Deleting argv[0] is prohibited"));
-			for (i = arg_no; i <= arg_end; i++) 
-				free(req->argv[i]);
-			i = arg_end - arg_no + 1;
-			memmove(req->argv + arg_no,
-				req->argv + arg_end + 1,
-				(req->argc - i)
-				  * sizeof(req->argv[0]));
-			req->argc -= i;
-			args_transformed = 1;
-			break;
-
-		case transform_setarg:
-			GET_TGT_VAL();
-			assign_string(target, val);
-			break;
-                }
-        }
+			if (node->pattern) {
+				mem = rush_expand_string(node->pattern, req);
+				val = mem;
+			} else {
+				mem = NULL;
+			}
+		}
+		res = transform_funtab[node->type].func(req, node,
+							val, target);
+		if (flags & XFORM_CHARGV)
+			args_transformed = res;
+		if (flags & XFORM_CMDLINE)
+			reparse_cmdline(req);
+		if (mem) {
+			free(mem);
+			mem = NULL;
+		}
+	}
 
         if (args_transformed) 
                 rebuild_cmdline(req);
@@ -668,7 +815,7 @@ run_transforms(struct rush_rule *rule, struct rush_request *req)
         }
 }
 
-void
+static void
 acct_on(struct rush_rule *rule, struct rush_request *req, pid_t pid)
 {
 	struct rush_wtmp wtmp;
@@ -684,7 +831,7 @@ acct_on(struct rush_rule *rule, struct rush_request *req, pid_t pid)
 		    RUSH_DB, strerror(errno));
 }
 
-void
+static void
 acct_off(void)
 {
 	if (rushdb_stop())
@@ -694,7 +841,7 @@ acct_off(void)
 	rushdb_close();
 }
 
-void
+static void
 fork_process(struct rush_rule *rule, struct rush_request *req)
 {
 	int status;
@@ -720,12 +867,12 @@ fork_process(struct rush_rule *rule, struct rush_request *req)
 
 	if (req->acct == rush_true)
 		acct_on(rule, req, pid);
-	debug1(2, _("Forked process %lu"), (unsigned long) pid);
+	debug(2, _("Forked process %lu"), (unsigned long) pid);
 	waitpid(pid, &status, 0);
 	if (WIFEXITED(status)) {
 		status = WEXITSTATUS(status);
-		debug2(2, _("%s: subprocess exited with code %d"),
-		       rule->tag, status);
+		debug(2, _("%s: subprocess exited with code %d"),
+		      rule->tag, status);
 	} else if (WIFSIGNALED(status)) {
 		logmsg(LOG_NOTICE, _("%s: subprocess terminated on signal %d"),
 		       rule->tag, WTERMSIG(status));
@@ -762,7 +909,7 @@ get_user_groups(struct rush_request *req, size_t *pgidc, gid_t **pgidv)
 	gidc = 1;
 
 	setgrent();
-	while (gr = getgrent()) {
+	while ((gr = getgrent())) {
 		char **p;
 		for (p = gr->gr_mem; *p; p++)
 			if (strcmp(*p, req->pw->pw_name) == 0) {
@@ -803,14 +950,14 @@ setowner(struct rush_request *req)
 		die(system_error, &req->i18n,
 		    _("seteuid(0) succeeded when it should not"));
 }
-
-void
+
+static void
 run_rule(struct rush_rule *rule, struct rush_request *req)
 {
         char **new_env;
 		
-        debug3(2, _("Rule %s at %s:%d matched"),
-	       rule->tag, rule->file, rule->line);
+        debug(2, _("Rule %s at %s:%d matched"),
+	      rule->tag, rule->file, rule->line);
 
         new_env = env_setup(rule->env);
         if (__debug_p(2)) {
@@ -832,7 +979,7 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 		const char *msg = rule->error_msg;
 		int custom = 1;
 		
-                debug1(2, _("Error message: %s"), msg);
+                debug(2, _("Error message: %s"), msg);
 		if (msg[0] == '@') {
 			int n;
 			
@@ -869,16 +1016,20 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
         run_transforms(rule, req);
 
         if (rule->chroot_dir) {
-                char *dir = expand_tilde(rule->chroot_dir,
-                                               req->pw->pw_dir);
-                debug1(2, _("Chroot dir: %s"), dir);
+                char *dir = expand_dir(rule->chroot_dir, req);
+                debug(2, _("Chroot dir: %s"), dir);
 		free(req->chroot_dir);
 		req->chroot_dir = dir;
         }
+        if (rule->home_dir) {
+                free(req->home_dir);
+                req->home_dir = expand_dir(rule->home_dir, req);
+		debug(2, _("Home dir: %s"), req->home_dir);
+	}
 
 	if (rule->gid != NO_GID) {
 		req->gid = rule->gid;
-		debug1(2, _("GID: %lu"), req->gid);
+		debug(2, _("GID: %lu"), (unsigned long) req->gid);
 	}
 	
 	if (rule->post_sockaddr.len)
@@ -917,24 +1068,26 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 		uid = req->pw->pw_uid;
 		pw = getpwuid(uid);
 		if (!pw)
-			die(nologin_error, NULL,
+			die(req->interactive ? nologin_error : usage_error,
+			    NULL,
 			    _("invalid uid %lu"), (unsigned long) uid);
 		req->pw = pw;
 	}
 
-        if (rule->home_dir) {
-                free(req->home_dir);
-                req->home_dir = expand_tilde(rule->home_dir, req->pw->pw_dir);
-		debug1(2, _("Home dir: %s"), req->home_dir);
+        if (req->home_dir) {
+		debug(2, _("chdir %s"), req->home_dir);
 		if (chdir(req->home_dir)) 
 			die(system_error, &req->i18n,
 			    _("cannot change to dir %s: %s"),
                     req->home_dir, strerror(errno));
 	}
 
-        debug2(2, _("Executing %s, %s"), PROGFILE(req), req->cmdline);
-	if (lint_option)
+        debug(2, _("Executing %s, %s"), PROGFILE(req), req->cmdline);
+	if (lint_option) {
+		if (dump_option)
+			dump_request(req, stdout);
 		exit(0);
+	}
 
 	if (req->fork == rush_true) 
 		fork_process(rule, req);
@@ -952,13 +1105,13 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 
 static char *command = NULL;
 static char *test_user_name = NULL;
+static int interactive;
 
 #include "rushopt.h"
 
 int
 main(int argc, char **argv)
 {
-        int rc;
         uid_t uid;
         struct rush_rule *rule;
         struct rush_request req;
@@ -981,35 +1134,33 @@ main(int argc, char **argv)
 		die(usage_error, NULL, _("invalid command line"));
 	
 	/* Relinquish root privileges in test mode */
-	if (lint_option)
-		setuid(getuid());
+	if (lint_option) {
+		if (setuid(getuid()))
+			die(system_error, NULL, "setuid: %s", strerror(errno));
+	}
 	
 	if (test_user_name) {
 		struct passwd *pw = getpwnam(test_user_name);
 		if (!pw)
 			die(usage_error, NULL, _("invalid user name"));
-		setreuid(pw->pw_uid, 0);
+		if (setreuid(pw->pw_uid, 0))
+			die(system_error, NULL, "setreuid: %s",
+			    strerror(errno));
 	}
 
         uid = getuid();
         if ((rush_pw = getpwuid(uid)) == NULL)
-                die(nologin_error, NULL,
+                die(system_error, NULL,
 		    _("invalid uid %lu"), (unsigned long) uid);
 
-        debug2(2, _("user %s, uid %lu"), rush_pw->pw_name,
-               (unsigned long) rush_pw->pw_uid);
+        debug(2, _("user %s, uid %lu"), rush_pw->pw_name,
+	      (unsigned long) rush_pw->pw_uid);
 
         parse_config();
 
 	if (!command) {
-		if (lint_option) 
+		if (lint_option && !interactive) 
 			exit(0);
-		else if (rush_interactive_shell) {
-			command = rush_interactive_shell;
-			debug1(2, _("Mapping interactive shell to \"%s\""),
-			       command);
-		} else 
-			die(usage_error, NULL, _("invalid command line"));
 	}
 
         if (__debug_p(2)) {
@@ -1023,7 +1174,20 @@ main(int argc, char **argv)
         }
 
 	memset(&req, 0, sizeof(req));
-        req.cmdline = xstrdup(command);
+	if (!command) {
+		req.interactive = 1;
+		command = "/bin/sh";
+	}
+
+	req.cmdline = xstrdup(command);
+
+	if (wordsplit(req.cmdline, &ws, WRDSF_DEFFLAGS))
+		die(system_error, NULL,
+		    _("wordsplit(%s) failed: %s"),
+		    req.cmdline, wordsplit_strerror(&ws));
+	wordsplit_get_words(&ws, &req.argc, &req.argv);
+	wordsplit_free(&ws);
+	
         req.pw = rush_pw;
 	req.umask = 022;
 	req.chroot_dir = NULL;
@@ -1031,30 +1195,26 @@ main(int argc, char **argv)
 	req.gid = NO_GID;
 	req.fork = rush_undefined;
 	req.acct = rush_undefined;
-
-	if (wordsplit(req.cmdline, &ws, WRDSF_DEFFLAGS))
-                die(system_error, NULL,
-                    _("wordsplit(%s) failed: %s"),
-                    req.cmdline, wordsplit_strerror(&ws));
-	req.argc = ws.ws_wordc;
-	req.argv = ws.ws_wordv;
-	ws.ws_wordc = 0;
-	ws.ws_wordv = NULL;
-
-	wordsplit_free(&ws);
 	
 	for (rule = rule_head; rule; rule = rule->next) {
-                rule = match_rule(rule, &req);
-                if (!rule)
-                        break;
-                if (debug_level) 
-                        logmsg(LOG_NOTICE,
-                               _("Serving request \"%s\" for %s by rule %s"),
-                               command, req.pw->pw_name, rule->tag);
+		if (req.interactive != rule->interactive)
+			continue;
+                if (run_tests(rule, &req))
+			continue;
+                if (debug_level) {
+			if (req.interactive)
+				logmsg(LOG_NOTICE,
+				       _("Serving interactive shell request for %s by rule %s"),
+				       req.pw->pw_name, rule->tag);
+			else
+				logmsg(LOG_NOTICE,
+				       _("Serving request \"%s\" for %s by rule %s"),
+				       command, req.pw->pw_name, rule->tag);
+		}
                 run_rule(rule, &req);
         }
-	die(usage_error, &req.i18n, 
+	die(req.interactive ? nologin_error : usage_error, &req.i18n,
 	    _("no matching rule for \"%s\", user %s"),
-	    req.cmdline, req.pw->pw_name);        
+	    req.cmdline, req.pw->pw_name);
         return 0;
 }
